class InvalidEventTypeDefinition(Exception):
    pass

class InvaildAuthkey(Exception):
    pass

class AccountNotFound(Exception):
    pass

class InvaildSession(Exception):
    pass

class UnauthorizedSession(Exception):
    pass

class UnknownTarget(Exception):
    pass

# FileNotFoundError

# PermissionError

class AccountMuted(Exception):
    pass

class TooLongMessage(Exception):
    pass

class InvaildArgument(Exception):
    pass

class NotSupportedVersion(Exception):
    pass

class DeprecatedImpl(Exception):
    pass

class EntangledSuperposition(Exception):
    pass

class MissingNecessaryOne(Exception):
    pass